#include "gtslink.h"

/************************************************************************\
*
* Module for single linked lists.
*
* $Ident: This file is part of the 'Gt-library' 1.0.0 $
* $Copyright: (C) 1997 by Gran Gustafsson $
*
\************************************************************************/

#ifndef NULL
#define NULL ((void *)0)
#endif

/************************************************************************\
*
* Setup list head.
*
\************************************************************************/
void GtSInit(GtSLink *sl)
{
	if(sl)
		sl->next=NULL;
}

/************************************************************************\
*
* Add link at the head of the list.
*
\************************************************************************/
void GtSAddHead(GtSLink *hd,GtSLink *sl)
{
	if(hd && sl)
	{
		sl->next=hd->next;
		hd->next=sl;
	}
}

/************************************************************************\
*
* Add link at the tail of the list.
*
\************************************************************************/
void GtSAddTail(GtSLink *hd,GtSLink *sl)
{
	if(hd && sl)
	{
		while(hd->next)
			hd=hd->next;
		sl->next=hd->next;
		hd->next=sl;
	}
}

/************************************************************************\
*
* Add link at given position or last.
*
\************************************************************************/
void GtSAddAt(GtSLink *hd,GtSLink *sl,unsigned long pos)
{
	if(hd && sl)
	{
		while(pos && hd->next)
		{
			pos--;
			hd=hd->next;
		}
		sl->next=hd->next;
		hd->next=sl;
	}
}

/************************************************************************\
*
* Remove link form list.
*
\************************************************************************/
void GtSRemove(GtSLink *hd,GtSLink *sl)
{
	if(hd && sl)
	{
		while(hd && (hd->next!=sl))
			hd=hd->next;
		if(hd)
			hd->next=sl->next;
		sl->next=NULL;
	}
}

/************************************************************************\
*
* Remove head of list.
*
\************************************************************************/
GtSLink *GtSRemHead(GtSLink *hd)
{
	GtSLink *sl=NULL;
	
	if(hd)
	{
		sl=hd->next;
		if(sl)
			hd->next=sl->next;
		else
			hd->next=NULL;
	}
	return(sl);
}

	
/************************************************************************\
*
* Get first link in list.
*
\************************************************************************/
GtSLink *GtSFirst(GtSLink *hd)
{
	if(hd)
		return(hd->next);
	else
		return(NULL);
}

/************************************************************************\
*
* Get next link in list.
*
\************************************************************************/
GtSLink *GtSNext(GtSLink *sl)
{
	if(sl)
		return(sl->next);
	else
		return(NULL);
}
