#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <proto.h>

#include "gtstrinput.h"
#include "gtapp.h"
#include "gterror.h"
#include "gtresname.h"
#include "gtcommon.h"

/************************************************************************\
*
* GtStrInput class.
*
* $Ident: This file is part of the 'Gt-library' 1.0.0 $
* $Copyright: (C) 1997 by Gran Gustafsson $
*
\************************************************************************/

#define GTSIF_PRESS		0x01
#define GTSIF_FOCUS		0x02
#define GTSIF_MARK		0x04
#define GTSIF_PASTE		0x08
#define GTSIF_CFILL		0x80

#define GTSIA_STRING	(GTDSA_MAX_NR)
#define GTSIA_ATOM_NR	(GTDSA_MAX_NR+1)

static Atom selatoms[GTSIA_ATOM_NR];

/************************************************************************\
*
* Create and init widget.
*
\************************************************************************/
GtStrInput *GtMkStrInput(char *n,GtManager *p,GtTag *tgs)
{
	GtStrInput *obj;
	
	if(!(obj=new GtStrInput()))
	{
		GtError(NULL,GTEC_NOOBJECT,"No GtStrInput object!\n");
		return(NULL);
	}
	if(!obj->Init(n,p,tgs))
	{
		GtDestroy(obj);
		obj=NULL;
	}
	return(obj);
}
	
/************************************************************************\
*
* Create and init widget.
*
\************************************************************************/
GtStrInput *GtMkStrInput(char *n,GtManager *p,ulong tag,...)
{
	va_list va;
	GtStrInput *obj;
	int ps;
	
	va_start(va,tag);
	ps=GtBeginVaTags();
	obj=GtMkStrInput(n,p,GtCopyVaTags(tag,&va));
	GtEndVaTags(ps);
	va_end(va);
	return(obj);
}
	
/************************************************************************\
*
* Setup empty strinput.
*
\************************************************************************/
GtStrInput::GtStrInput(void)
{
	str=NULL;
	ins=NULL;
	mark=NULL;
	maxl=40;
	len=0;
	pos=0;
	first=0;
	siflags=GTSIF_CFILL;
	ibw=0;
	bdrw=1;
	ibdrw=1;
	paste=0;
	ftype=GTF_DBL_OUT;
}

/************************************************************************\
*
* Destroy strinput.
*
\************************************************************************/
void GtStrInput::Destroy(void)
{
	if(mark)
	{
		dsp->DisOwnSelection(this,XA_PRIMARY);
		free(mark);
	}
	if(str)
		free(str);
	GtWidget::Destroy();
}

/************************************************************************\
*
* Setup strinput.
*
\************************************************************************/
sbool GtStrInput::Init(char *n,GtManager *p,GtTag *tgs)
{
	GtTag *t;
	char *istr=NULL;
	XCharStruct xcs;
	word mxs,mys;
	sbool cf=TRUE;
	
	if(!GtWidget::Init(n,p,tgs))
		return(FALSE);
	app->GetRes(this,GTR_STRING,GTRT_STRING,&istr);
	app->GetRes(this,GTR_FRAME,GTRT_FRAME,&ftype);
	app->GetRes(this,GTR_IBDR,GTRT_BYTE,&ibdrw);
	app->GetRes(this,GTR_BDR,GTRT_BYTE,&bdrw);
	app->GetRes(this,GTR_CFILL,GTRT_BOOL,&cf);
	while((t=GtNextTag(&tgs)))
	{
		switch(t->tag)
		{
			case GT_String:
				istr=(char *)t->d.ptr;
				break;
			case GT_MaxLength:
				if(t->d.val)
					maxl=t->d.val;
				break;
			case GT_InnerBorder:
				ibdrw=t->d.val;
				break;
			case GT_Border:
				bdrw=t->d.val;
				break;
			case GT_FrameType:
				ftype=t->d.val;
				if(ftype>GTF_LAST_NR)
					ftype=GTF_DBL_OUT;
				break;
			case GT_BlockCursor:
				if(t->d.val)
					cf=TRUE;
				else
					cf=FALSE;
				break;
		}
	}
	if(cf)
		siflags|=GTSIF_CFILL;
	else
		siflags&=~GTSIF_CFILL;
	memcpy(selatoms,dsp->GetStdAtoms(),GTDSA_MAX_NR*sizeof(Atom));
	selatoms[GTSIA_STRING]=XA_STRING;
	switch(ftype)
	{
		case GTF_NONE:
		case GTF_SIMPLE:
		case GTF_IN:
		case GTF_OUT:
			ibw=ibdrw+bdrw;
			break;
		case GTF_DBL_IN:
		case GTF_DBL_OUT:
			ibw=ibdrw+2*bdrw;
			break;
	}
	if(!this->PrivateGC(GT_Foreground))
		return(FALSE);
	gtfnt->MaxExtents(&xcs);
	foff=xcs.ascent;
	fh=foff+xcs.descent;
	mxs=GTMAX(xcs.width+2*ibw,2*ibw+1);
	mys=GTMAX(fh+2*ibw,2*ibw+1);
	gtfnt->MinExtents(&xcs);
	mw=xcs.width;
	if(!(str=(char *)malloc((maxl+1)*2)))
	{
		GtError(this,GTEC_NOMEMORY,"No memory for string buffers!\n");
		return(FALSE);
	}
	memset(str,0,(maxl+1)*2);
	ins=str+maxl+1;
	if(istr)
	{
		InsertStr(istr);
		MoveCurs(0);
	}
	if(!(paste=XInternAtom(dsp->XDsp(),"GT_PASTE_STRING",False)))
	{
		GtError(this,GTEC_NOXRESOURCE,"No atom for string pasting!\n");
		return(FALSE);
	}
	this->SetMinMaxSize(mxs,mys,GTS_MAX,GTS_MAX);
	return(TRUE);
}

/************************************************************************\
*
* Set strinput attributes.
*
\************************************************************************/
sbool GtStrInput::Set(GtTag *tgs)
{
	GtTag *t;
	char *istr=NULL;
	sbool fch=FALSE,lch=FALSE,vch=FALSE;
	XCharStruct xcs;
	word mxs=1,mys=1;
	
	if(!GtWidget::Set(tgs))
		return(FALSE);
	while((t=GtNextTag(&tgs)))
	{
		switch(t->tag)
		{
			case GT_String:
				istr=(char *)t->d.ptr;
				break;
			case GT_MaxLength:
				if(t->d.val)
					maxl=t->d.val;
				lch=TRUE;
				break;
			case GT_CursPosition:
				MoveCurs(t->d.val);
				break;
			case GT_InnerBorder:
				ibdrw=t->d.val;
				vch=TRUE;
				break;
			case GT_Border:
				bdrw=t->d.val;
				vch=TRUE;
				break;
			case GT_FrameType:
				ftype=t->d.val;
				if(ftype>GTF_LAST_NR)
					ftype=GTF_DBL_OUT;
				vch=TRUE;
				break;
			case GT_Font:
			case GT_FontList:
				fch=TRUE;
				break;
		}
	}
	if(vch)
	{
		switch(ftype)
		{
			case GTF_NONE:
			case GTF_SIMPLE:
			case GTF_IN:
			case GTF_OUT:
				ibw=ibdrw+bdrw;
				break;
			case GTF_DBL_IN:
			case GTF_DBL_OUT:
				ibw=ibdrw+2*bdrw;
				break;
		}
	}
	if(fch)
	{
		gtfnt->MaxExtents(&xcs);
		foff=xcs.ascent;
		fh=foff+xcs.descent;
		mxs=GTMAX(xcs.width+2*ibw,2*ibw+1);
		mys=GTMAX(fh+2*ibw,2*ibw+1);
		gtfnt->MinExtents(&xcs);
		mw=xcs.width;
		mvis=(xsize-2*ibw)/mw;
	}
	if(lch)
	{
		if(str)
			free(str);
		if(!(str=(char *)malloc((maxl+1)*2)))
		{
			GtError(this,GTEC_NOMEMORY,"No memory for string buffers!\n");
			return(FALSE);
		}
		memset(str,0,(maxl+1)*2);
		ins=str+maxl+1;
	}
	if(istr)
	{
		this->MoveCurs(0);
		this->DeleteStr(len);
		this->InsertStr(istr);
	}
	if(fch)
		this->SetMinMaxSize(mxs,mys,GTS_MAX,GTS_MAX);
	if(lch || istr || vch)
		this->Refresh();
	return(TRUE);
}
	
/************************************************************************\
*
* Get strinput attributes.
*
\************************************************************************/
void GtStrInput::Get(GtTag *tgs)
{
	GtTag *t;
	long *ptr;
	
	GtWidget::Get(tgs);
	while((t=GtNextTag(&tgs)))
	{
		if(!(ptr=(long *)t->d.ptr))
			continue;
		switch(t->tag)
		{
			case GT_String:
				*ptr=(long)str;
				break;
			case GT_MaxLength:
				*ptr=maxl;
				break;
			case GT_CursPosition:
				*ptr=pos;
				break;
			case GT_InnerBorder:
				*ptr=ibdrw;
				break;
			case GT_Border:
				*ptr=bdrw;
				break;
			case GT_FrameType:
				*ptr=ftype;
				break;
		}
	}
}

/************************************************************************\
*
* Check if class is (sub)class of named class.
*
\************************************************************************/
sbool GtStrInput::IsSubClassOf(char *cn)
{
	if(strcmp(cn,"GtStrInput")==0)
		return(TRUE);
	else
		return(GtWidget::IsSubClassOf(cn));
}

/************************************************************************\
*
* Handle all key events if in focus.
*
\************************************************************************/
int GtStrInput::Handle(XEvent *xe)
{
	GtSIDoneCD cd;
	int ret;
	
	ret=GtWidget::Handle(xe);
	if((ret==GTHR_SEARCH) && (siflags & GTSIF_FOCUS))
	{
		switch(xe->type)
		{
			case KeyPress:
				this->Key(xe);
				break;
			case ButtonPress:
				siflags&=~GTSIF_MARK;
				dsp->ClrFocus(this);
				cd.why=GTDW_MOUSE;
				cd.str=str;
				cd.len=len;
				done.CallCB(this,&cd);
				/* OBS !! Falling through to default case */
			default:
				return(ret);
		}
		return(GTHR_USED);
	}
	return(ret);
}

/************************************************************************\
*
* Set strinput event mask.
*
\************************************************************************/
void GtStrInput::SetEMask(ulong em)
{
	GtWidget::SetEMask(em|ExposureMask|KeyPressMask|ButtonPressMask|
												ButtonReleaseMask|Button1MotionMask);
}

/************************************************************************\
*
* Check for focus notifies.
*
\************************************************************************/
void GtStrInput::Notify(ulong tag)
{
	GtWidget::Notify(tag);
	switch(tag)
	{
		case GT_FocusIn:
			siflags|=GTSIF_FOCUS;
			this->Refresh();
			break;
		case GT_FocusOut:
			siflags&=~GTSIF_FOCUS;
			this->Refresh();
			break;
	}
}

/************************************************************************\
*
* Add strinput callbacks.
*
\************************************************************************/
void GtStrInput::AddCB(ulong tag,GtCallB cb,GtUDPtr ud)
{
	switch(tag)
	{
		case GT_Edit:
			edit.AddCB(cb,ud);
			break;
		case GT_Done:
			done.AddCB(cb,ud);
			break;
		default:
			GtWidget::AddCB(tag,cb,ud);
			break;
	}
}

/************************************************************************\
*
* Remove strinput callbacks.
*
\************************************************************************/
void GtStrInput::RemCB(ulong tag,GtCallB cb)
{
	switch(tag)
	{
		case GT_Edit:
			edit.RemCB(cb);
			break;
		case GT_Done:
			done.RemCB(cb);
			break;
		default:
			GtWidget::RemCB(tag,cb);
			break;
	}
}

/************************************************************************\
*
* Set cursor position.
*
\************************************************************************/
void GtStrInput::SetCursor(ulong npos)
{
	GtSIEditCD cd;
	
	cd.why=GTEW_MOVE;
	cd.xevent=NULL;
	cd.key=NoSymbol;
	cd.state=0;
	cd.currpos=pos;
	cd.newpos=npos;
	cd.ins=ins;
	cd.len=0;
	cd.doedit=TRUE;
	edit.CallCB(this,&cd);
	if(cd.doedit)
	{
		if(siflags & GTSIF_MARK)
		{
			siflags&=~GTSIF_MARK;
			this->ReDraw(mstart,mend);
		}
		this->MoveCurs(cd.newpos);
	}
}

/************************************************************************\
*
* Insert string at current position.
*
\************************************************************************/
void GtStrInput::InsertString(char *istr)
{
	GtSIEditCD cd;
	
	strncpy(ins,istr,maxl);
	cd.why=GTEW_PASTE;
	cd.xevent=NULL;
	cd.key=NoSymbol;
	cd.state=0;
	cd.currpos=pos;
	cd.newpos=pos+strlen(ins);
	cd.ins=ins;
	cd.len=strlen(ins);
	cd.doedit=TRUE;
	edit.CallCB(this,&cd);
	if(cd.doedit)
	{
		if(siflags & GTSIF_MARK)
		{
			siflags&=~GTSIF_MARK;
			this->ReDraw(mstart,mend);
		}
		this->InsertStr(cd.ins);
	}
}
	
/************************************************************************\
*
* Replace string.
*
\************************************************************************/
void GtStrInput::ReplaceString(char *rstr)
{
	GtSIEditCD cd;
	
	strncpy(ins,rstr,maxl);
	cd.why=GTEW_RPLACE;
	cd.xevent=NULL;
	cd.key=NoSymbol;
	cd.state=0;
	cd.currpos=pos;
	cd.newpos=pos+strlen(ins);
	cd.ins=ins;
	cd.len=strlen(ins);
	cd.doedit=TRUE;
	edit.CallCB(this,&cd);
	if(cd.doedit)
	{
		if(siflags & GTSIF_MARK)
		{
			siflags&=~GTSIF_MARK;
			this->ReDraw(mstart,mend);
		}
		this->MoveCurs(0);
		this->DeleteStr(len);
		this->InsertStr(cd.ins);
	}
}

/************************************************************************\
*
* Delete string at current position.
*
\************************************************************************/
void GtStrInput::DeleteString(ulong dlen)
{
	GtSIEditCD cd;
	
	cd.why=GTEW_DELETE;
	cd.xevent=NULL;
	cd.key=NoSymbol;
	cd.state=0;
	cd.currpos=pos;
	cd.newpos=pos;
	cd.ins=ins;
	cd.len=0;
	cd.doedit=TRUE;
	edit.CallCB(this,&cd);
	if(cd.doedit)
	{
		if(siflags & GTSIF_MARK)
		{
			siflags&=~GTSIF_MARK;
			this->ReDraw(mstart,mend);
		}
		this->DeleteStr(dlen);
	}
}

/************************************************************************\
*
* Mark part of the string.
*
\************************************************************************/
void GtStrInput::MarkString(ulong start,ulong end)
{
	if(siflags & GTSIF_FOCUS)
	{
		mstart=GTMIN(start,end);
		mend=GTMIN(GTMAX(start,end),len);
		siflags|=GTSIF_MARK;
		last=CurrentTime;
		this->AllocMark();
		this->ReDraw(mstart,mend);
	}
}

/************************************************************************\
*
* Activate strinput.
*
\************************************************************************/
void GtStrInput::Activate(void)
{
	if(!(flags & GTWF_DISABLED))
	{
		siflags&=~GTSIF_MARK;
		dsp->SetFocus(this);
	}
}

/************************************************************************\
*
* Deactivate strinput.
*
\************************************************************************/
void GtStrInput::Deactivate(void)
{
	siflags&=~GTSIF_MARK;
	dsp->ClrFocus(this);
}

/************************************************************************\
*
* Set size and position of strinput.
*
\************************************************************************/
void GtStrInput::SetPosSize(word xp,word yp,word xs,word ys)
{
	int cxs;
	
	GtWidget::SetPosSize(xp,yp,xs,ys);
	rect.x=ibw;
	rect.y=ibw;
	rect.width=xsize-2*ibw;
	rect.height=ysize-2*ibw;
	if(gfx)
		gtfore->SetClipRectangles(0,0,&rect,1,YXBanded);
	mvis=(xsize-2*ibw)/mw;
	if(siflags & GTSIF_FOCUS)
	{
		cxs=gtfnt->StrWidth(&str[first],pos-first+1);
		if(cxs>xsize-2*ibw)
		{
			while(cxs>xsize-2*ibw)
			{
				cxs-=gtfnt->CharWidth(str[first]);
				first++;
			}
		}
	}
}

/************************************************************************\
*
* Draw strinput.
*
\************************************************************************/
sbool GtStrInput::Draw(XEvent *xe)
{
	int n;
	
	if(xe && ((xe->type!=Expose) || (xe->xexpose.count!=0)))
		return(TRUE);
	this->ReDraw(first,len);
	if(flags & GTWF_DISABLED)
		gtdis->DrawRectangle(gtwin,0,0,xsize,ysize,TRUE);
	switch(ftype)
	{
		case GTF_SIMPLE:
			for(n=0; n<bdrw; n++)
				gtfore->DrawRectangle(gtwin,n,n,xsize-2*n-1,ysize-2*n-1,FALSE);
			break;
		case GTF_DBL_IN:
			for(n=0; n<bdrw; n++)
			{
				GT3DBOX(gtdrk,gtlht,gtwin,n,n,xsize-2*n,ysize-2*n);
				GT3DBOX(gtlht,gtdrk,gtwin,bdrw+n,bdrw+n,xsize-2*(bdrw+n),ysize-2*(bdrw+n));
			}
			break;
		case GTF_DBL_OUT:
			for(n=0; n<bdrw; n++)
			{
				GT3DBOX(gtlht,gtdrk,gtwin,n,n,xsize-2*n,ysize-2*n);
				GT3DBOX(gtdrk,gtlht,gtwin,bdrw+n,bdrw+n,xsize-2*(bdrw+n),ysize-2*(bdrw+n));
			}
			break;
		case GTF_IN:
			for(n=0; n<bdrw; n++)
			{
				GT3DBOX(gtdrk,gtlht,gtwin,n,n,xsize-2*n,ysize-2*n);
			}
			break;
		case GTF_OUT:
			for(n=0; n<bdrw; n++)
			{
				GT3DBOX(gtlht,gtdrk,gtwin,n,n,xsize-2*n,ysize-2*n);
			}
			break;			
	}
	return(TRUE);
}

/************************************************************************\
*
* Key strinput events.
*
\************************************************************************/
sbool GtStrInput::Key(XEvent *xe)
{
	GtSIEditCD cd;
	KeySym ks;
	ulong np;
	int l;
	
	if(!(siflags & GTSIF_FOCUS))
		return(FALSE);
	if(xe->type==KeyPress)
	{
		l=XLookupString(&xe->xkey,ins,maxl,&ks,NULL);
		ins[l]='\0';
		np=pos;
		if(IsModifierKey(ks))
			return(TRUE);
		switch(ks)
		{
			case XK_Left:
				cd.why=GTEW_MOVE;
				if(pos>0)
					np=pos-1;
				break;
			case XK_Right:
				cd.why=GTEW_MOVE;
				if(pos<len)
					np=pos+1;
				break;
			case XK_Up:
			case XK_Down:
			case XK_Return:
			case XK_Tab:
				cd.why=GTEW_DONE;
				break;
			case XK_Home:
				cd.why=GTEW_MOVE;
				np=0;
				break;
			case XK_End:
				cd.why=GTEW_MOVE;
				np=len;
				break;
			case XK_Delete:
				cd.why=GTEW_DELETE;
				break;
			case XK_BackSpace:
				cd.why=GTEW_BSPACE;
				if(pos>0)
					np=pos-1;
				break;
			default:
				cd.why=GTEW_TYPED;
				np=pos+GTMIN((ulong)l,maxl-len);
				break;
		}
		cd.xevent=xe;
		cd.key=ks;
		cd.state=xe->xkey.state;
		cd.currpos=pos;
		cd.newpos=np;
		cd.ins=ins;
		cd.len=l;
		cd.doedit=TRUE;
		edit.CallCB(this,&cd);
		if(!cd.doedit)
			return(TRUE);
		this->DoKey(&cd);
	}
	return(TRUE);
}
				
/************************************************************************\
*
* Mouse strinput events.
*
\************************************************************************/
sbool GtStrInput::Mouse(XEvent *xe)
{
	ulong n;
	int xp,yp,cxp,cxs;
	sbool found=FALSE;
	
	switch(xe->type)
	{
		case ButtonPress:
			if(flags & GTWF_DISABLED)
				break;
			if((xe->xbutton.button==Button1) || (xe->xbutton.button==Button2))
			{
				if(((xe->xbutton.button==Button1) && (siflags & GTSIF_PASTE)) ||
					((xe->xbutton.button==Button2) && (siflags & GTSIF_PASTE)))
					return(TRUE);
				xp=xe->xbutton.x;
				yp=xe->xbutton.y;
				if((xp<ibw) || (xp>=xsize-ibw) || (yp<ibw) || (yp>=ysize-ibw))
					return(TRUE);
				cxp=ibw;
				for(n=first; (n<len) && (n<first+mvis); n++)
				{
					cxs=gtfnt->CharWidth(str[n]);
					if(cxs>xsize-2*ibw)
						break;
					if((xp>=cxp) && (xp<cxp+cxs))
						break;
					cxp+=cxs;
				}
				pos=n;
				siflags&=~GTSIF_MARK;
				if(xe->xbutton.button==Button2)
				{
					siflags&=~GTSIF_MARK;
					this->ReDraw(mstart,mend);
					siflags|=GTSIF_PASTE;
					dsp->GetSelection(this,XA_PRIMARY,XA_STRING,paste,xe->xbutton.time,
														GtStrInput::Paste,this);
					return(TRUE);
				}
				if(siflags & GTSIF_FOCUS)
				{
					this->Refresh();
				}
				else
					dsp->SetFocus(this);
				if(siflags & GTSIF_FOCUS)
					siflags|=GTSIF_PRESS;
				return(TRUE);
			}
			break;
		case ButtonRelease:
			if(xe->xbutton.button==Button1)
			{
				siflags&=~GTSIF_PRESS;
				last=xe->xbutton.time;
				this->AllocMark();
				this->ReDraw(mstart,mend);
				return(TRUE);
			}
			break;
		case MotionNotify:
			if((xe->xmotion.state & Button1Mask) && (siflags & GTSIF_PRESS))
			{
				xp=xe->xmotion.x;
				yp=xe->xmotion.y;
				if((xp<ibw) || (xp>=xsize-ibw) || (yp<ibw) || (yp>=ysize-ibw))
					return(TRUE);
				cxp=ibw;
				for(n=first; (n<len) && (n<first+mvis); n++)
				{
					cxs=gtfnt->CharWidth(str[n]);
					if(cxs>xsize-2*ibw)
						break;
					if((xp>=cxp) && (xp<cxp+cxs))
					{
						mstart=GTMIN(n,pos);
						mend=GTMAX(n,pos);
						found=TRUE;
						break;
					}
					cxp+=cxs;
				}
				if(!found)
					return(TRUE);
				if(mend-mstart)
					siflags|=GTSIF_MARK;
				this->Refresh();
				return(TRUE);
			}
			break;
	}
	return(FALSE);
}

/************************************************************************\
*
* Do edit after keypress.
*
\************************************************************************/
void GtStrInput::DoKey(GtSIEditCD *cd)
{
	GtSIDoneCD cd2;
	ulong op;
	
	switch(cd->why)
	{
		case GTEW_MOVE:
			if(!(cd->state & ShiftMask) && (siflags & GTSIF_MARK))
			{
				siflags&=~GTSIF_MARK;
				this->ReDraw(mstart,mend);
			}
			op=pos;
			this->MoveCurs(cd->newpos);
			if(cd->state & ShiftMask)
			{
				if(!(siflags & GTSIF_MARK))
				{
					siflags|=GTSIF_MARK;
					mstart=op;
					mend=op;
				}
				mstart=GTMIN(mstart,pos);
				mend=GTMAX(mend,pos);
				this->AllocMark();
				this->ReDraw(mstart,mend);
			}
			break;
		case GTEW_DELETE:
			if(siflags & GTSIF_MARK)
			{
				siflags&=~GTSIF_MARK;
				this->MoveCurs(mstart);
				this->DeleteStr(mend-mstart+1);
			}
			else
			{
				this->MoveCurs(cd->newpos);
				this->DeleteStr(1);
			}
			break;
		case GTEW_BSPACE:
			if(siflags & GTSIF_MARK)
			{
				siflags&=~GTSIF_MARK;
				this->MoveCurs(mstart);
				this->DeleteStr(mend-mstart+1);
			}
			else
			{
				this->MoveCurs(cd->newpos);
				this->DeleteStr(cd->currpos<cd->newpos ? cd->newpos-cd->currpos
																							 : cd->currpos-cd->newpos);
			}
			break;
		case GTEW_DONE:
			switch(cd->key)
			{
				case XK_Up:
					cd2.why=GTDW_UP;
					break;
				case XK_Down:
					cd2.why=GTDW_DOWN;
					break;
				case XK_Return:
					cd2.why=GTDW_ENTER;
					break;
				case XK_Tab:
					if(cd->xevent->xkey.state & ShiftMask)
						cd2.why=GTDW_UPTAB;
					else
						cd2.why=GTDW_DOWNTAB;
					break;
			}
			siflags&=~GTSIF_MARK;
			dsp->ClrFocus(this);
			cd2.str=str;
			cd2.len=len;
			done.CallCB(this,&cd2);
			break;
		case GTEW_TYPED:
			if(siflags & GTSIF_MARK)
			{
				siflags&=~GTSIF_MARK;
				this->MoveCurs(mstart);
				this->DeleteStr(mend-mstart+1);
			}
			this->InsertStr(cd->ins);
			break;
	}
}

/************************************************************************\
*
* Move cursor to new position.
*
\************************************************************************/
void GtStrInput::MoveCurs(ulong npos)
{
	int xs;
	ulong op;

	if(npos>len)
		return;
	if(npos<first)
	{
		first=npos;
		pos=npos;
		this->ReDraw(pos,len);
	}
	else if(npos>pos)
	{
		xs=gtfnt->StrWidth(&str[first],npos-first+1);
		if(xs>xsize-2*ibw)
		{
			while(xs>xsize-2*ibw)
			{
				xs-=gtfnt->CharWidth(str[first]);
				first++;
			}
			pos=npos;
			this->ReDraw(first,len);
		}
		else
		{
			op=pos;
			pos=npos;
			this->ReDraw(op,op);
			this->ReDraw(pos,pos);
		}
	}
	else
	{
		op=pos;
		pos=npos;
		this->ReDraw(op,op);
		this->ReDraw(pos,pos);
	}
}

/************************************************************************\
*
* Insert string into strinput.
*
\************************************************************************/
void GtStrInput::InsertStr(char *s)
{
	int l,n,i,xs;
	ulong op;
	
	l=strlen(s);
	l=GTMIN((ulong)l,maxl-len);
	if(!l)
		return;
	i=len+l;
	for(n=len-pos+1; n; n--)
	{
		str[i]=str[i-l];
		i--;
	}
	for(n=0; n<l; n++)
	{
		str[pos+n]=s[n];
	}
	op=pos;
	len+=l;
	pos+=l;
	xs=gtfnt->StrWidth(&str[first],pos-first+1);
	if((siflags & GTSIF_FOCUS) && (xs>xsize-2*ibw))
	{
		while(xs>xsize-2*ibw)
		{
			xs-=gtfnt->CharWidth(str[first]);
			first++;
		}
		this->ReDraw(first,len);
	}
	else
		this->ReDraw(op,len);
}

/************************************************************************\
*
* Delete string from strinput.
*
\************************************************************************/
void GtStrInput::DeleteStr(ulong dl)
{
	ulong n;
	
	dl=GTMIN(dl,len-pos);
	if(!dl)
		return;
	for(n=0; n<len; n++)
	{
		str[pos+n]=str[pos+n+dl];
		if(!str[pos+n])
			break;
	}
	len-=dl;
	this->ReDraw(pos,len);
}

/************************************************************************\
*
* Redraw part of string.
*
\************************************************************************/
void GtStrInput::ReDraw(ulong fst,ulong lst)
{
	int xp=0,xs,mxp,mxs;
	ulong l;
	
	// Check if redraw is visible
	if((lst<fst) || (lst<first) || (fst>first+mvis))
		return;
	// Redraw from first char
	if(fst<first)
		fst=first;
	// Redraw no more than maximum visible or length
	if((lst>first+mvis) || (lst>len))
		lst=GTMIN(first+mvis,len);
	// Get offset from first char
	if(fst-first)
		xp=gtfnt->StrWidth(&str[first],fst-first);
	// Check if offset is past end
	if(xp>xsize-2*ibw)
		return;
	// Get size of string
	xs=gtfnt->StrWidth(&str[fst],lst-fst+1);
	// Make sure it is not to long
	xs=GTMIN(xs,xsize-2*ibw-xp);
	// Clear string area
	if(xs)
		gtwin->Clear(ibw+xp,ibw,xs,ysize-2*ibw);
	// Clear after string (for delete)
	if((lst==len) && (xsize-2*ibw-xs-xp>0))
		gtwin->Clear(ibw+xp+xs,ibw,xsize-2*ibw-xs-xp,ysize-2*ibw);
	// Check if any string part marked
	if((siflags & GTSIF_MARK) && (mstart<=lst) && (mend>=fst))
	{
		mxp=xp;
		if(mstart>fst)
		{
			mxp+=gtfnt->StrWidth(&str[fst],mstart-fst);
			mxs=gtfnt->StrWidth(&str[mstart],mend-mstart+1);
		}
		else
		{
			mxs=gtfnt->StrWidth(&str[fst],mend-fst+1);
		}
		mxs=GTMIN(xs-(mxp-xp),mxs);
		gtlht->DrawRectangle(gtwin,ibw+mxp,ibw,mxs,ysize-2*ibw,TRUE);
	}
	// Check if cursor visible
	if((siflags & GTSIF_FOCUS) && (pos<=lst) && (pos>=fst))
	{
		mxp=xp;
		if(pos>fst)
			mxp+=gtfnt->StrWidth(&str[fst],pos-fst);
		mxs=gtfnt->CharWidth(str[pos] ? str[pos] : ' ');
		if(siflags & GTSIF_CFILL)
			gtfill->DrawRectangle(gtwin,ibw+mxp,ibw,mxs,ysize-2*ibw,TRUE);
		else
			gtfill->DrawLine(gtwin,ibw+mxp,ibw,ibw+mxp,ysize-ibw-1);
	}
	// Draw string
	if((l=GTMIN(GTMIN(lst-fst+1,mvis),len-fst)))
		gtfore->DrawString(gtwin,ibw+xp,ibw+foff,&str[fst],l,FALSE);
}

/************************************************************************\
*
* Allocate marked string and copy it.
*
\************************************************************************/
void GtStrInput::AllocMark(void)
{
	ulong l;
	
	if(siflags & GTSIF_MARK)
	{
		if(mark)
			free(mark);
		mark=NULL;
		l=mend-mstart+1;
		if(!(mark=(char *)malloc(l+1)))
			return;
		strncpy(mark,&str[mstart],l);
		mark[l]='\0';
		if(!dsp->OwnSelection(this,XA_PRIMARY,last,GtStrInput::MarkConv,
				GtStrInput::MarkLose,GtStrInput::MarkDone,this))
		{
			free(mark);
			siflags&=~GTSIF_MARK;
		}
	}
}

/************************************************************************\
*
* Free marked string.
*
\************************************************************************/
void GtStrInput::FreeMark(void)
{
	if(mark)
		free(mark);
	mark=NULL;
}

/************************************************************************\
*
* Convert marked string.
*
\************************************************************************/
int GtStrInput::MarkConv(GtBase *obj,GtCDPtr cd,GtUDPtr ud)
{
	GtStrInput *si;
	GtSelReqCD *scd;
	
	obj=obj;
	scd=(GtSelReqCD *)cd;
	si=(GtStrInput *)ud;
	if(scd->target==selatoms[GTDSA_TARGETS])
	{
		scd->target=XA_ATOM;
		scd->fmt=32;
		scd->data=(ubyte *)selatoms;
		scd->len=GTSIA_ATOM_NR;
	}
	else if(scd->target==XA_STRING)
	{
		if(si->mark)
		{
			scd->target=XA_STRING;
			scd->fmt=8;
			scd->data=(ubyte *)si->mark;
			scd->len=strlen(si->mark);
		}
		else
		{
			scd->ok=FALSE;
		}
	}
	return(GTCB_CONT);
}

/************************************************************************\
*
* Lose selection.
*
\************************************************************************/
int GtStrInput::MarkLose(GtBase *obj,GtCDPtr cd,GtUDPtr ud)
{
	GtStrInput *si;
	
	obj=obj;
	cd=cd;
	si=(GtStrInput *)ud;
	si->siflags&=~GTSIF_MARK;
	si->ReDraw(si->mstart,si->mend);
	return(GTCB_CONT);
}

/************************************************************************\
*
* Done with selection.
*
\************************************************************************/
int GtStrInput::MarkDone(GtBase *obj,GtCDPtr cd,GtUDPtr ud)
{
	GtStrInput *si;

	obj=obj;
	cd=cd;
	si=(GtStrInput *)ud;
	si->FreeMark(); 
	return(GTCB_CONT);
}

/************************************************************************\
*
* Paste string in widget.
*
\************************************************************************/
int GtStrInput::Paste(GtBase *obj,GtCDPtr cd,GtUDPtr ud)
{
	GtStrInput *si;
	GtSelGetCD *scd;
	
	obj=obj;
	scd=(GtSelGetCD *)cd;
	si=(GtStrInput *)ud;
	if((scd->type==XA_STRING) && scd->data && scd->len)
		si->InsertString((char *)scd->data);
	si->siflags&=~GTSIF_PASTE;
	si->dsp->SetFocus(si);
	return(GTCB_CONT);
}
